# -*- coding: utf-8 -*-
# Importing os module for system operative utilities
import os
# Importing SQLContext:
from pyspark.sql.session import SparkSession
# Importando modulo para eliminar carpetas
from shutil import rmtree
# Importando modulo para trabajar con urls
import urllib.request
# Importando modulo de expresiones regulares
import re
# Import SparkContext
import pyspark

class Utilites():
    def __init__(self):
        """

        :param SparkContext:
         A Spark context is essentially a client of Spark’s execution environment and acts as the master of your
         Spark application (don’t get confused with the other meaning of Master in Spark, though).
        """

        # Setting SQLContext as a global variable of the class
        self.spark = SparkSession.builder.enableHiveSupport().getOrCreate()
        # Setting SparkContent as a global variable of the class
        self.__sc = self.spark.sparkContext

    def readDatasetCsv(self, path, delimiterMark=';', header='true'):
        """This funcion read a dataset from a csv file.

        :param path     Path or location of the file.
        :param delimiterMark   Usually delimiter mark are ',' or ';'.
        :param  header:     Tell the function is the dataset has a header row.

        :return dataFrame
        """
        assert ((header == 'true') or (header == 'false')), "Error, header argument must be 'true' or 'false'. " \
                                                             "header must be string dataType, i.e: header='true'"
        assert isinstance(delimiterMark, str), "Error, delimiter mark argumen must be string dataType."

        assert isinstance(path, str), "Error, path argument must be string datatype."

        return self.spark.read \
                .format('csv') \
                .options(header=header) \
                .options(delimiter=delimiterMark) \
                .options(inferSchema='true') \
                .load(path)

    @classmethod
    def readDatasetParquet(self, path):
        """This function allows user to read parquet files. It is import to clarify that this method is just based
        on the sqlContext.read.parquet(path) Apache Spark method. Only assertion instructions has been added to
        ensure user has more hints about what happened when something goes wrong.
        :param  path    Path or location of the file. Must be string dataType.

        :return dataFrame"""
        assert isinstance(path, str), "Error: path argument must be string dataType."
        assert (("file:///" == path[0:8]) or ("hdfs:///" == path[0:8])), "Error: path must be with a 'file://' prefix \
        if the file is in the local disk or a 'path://' prefix if the file is in the Hadood file system"
        return self.spark.read.parquet(path)


    def csvToParquet(self, inputPath, outputPath, delimiterMarkCsv, headerCsv, numPartitions=None):
        """This method transform a csv dataset file into a parquet.

        The method reads a existing csv file using the inputPath, delimiterMarkCsv and headerCsv arguments.

        :param  inputPath   Address location of the csv file.
        :param  outputPath  Address where the new parquet file will be stored.
        :param  delimiterMarkCsv    Delimiter mark of the csv file, usually is ',' or ';'.
        :param  headerCsv   This argument specifies if csv file has header or not.
        :param  numParitions Specifies the number of partitions the user wants to write the dataset."""

        df = self.readDatasetCsv(inputPath, delimiterMarkCsv, 'true')

        if numPartitions != None:
            assert (numPartitions <= df.rdd.getNumPartitions()), "Error: numPartitions specified is greater that the" \
                                                                "partitions in file store in memory."
            # Writting dataset:
            df.coalesce(numPartitions).write.parquet(outputPath)

        else:
            df.write.parquet(outputPath)

    def setCheckPointFolder(self, path, fileSystem):
        """Function that receives a workspace path where a folder is created.
        This folder will store temporal
        dataframes when user writes the DataFrameTransformer.checkPoint().

        This function needs the sc parameter, which is the spark context in order to
        tell spark where is going to save the temporal files.

        It is recommended that users deleted this folder after all transformations are completed
        and the final dataframe have been saved. This can be done with deletedCheckPointFolder function.

        :param path     Location of the dataset (string).
        :param fileSystem   Describes if file system is local or hadoop file system.

        """

        assert (isinstance(fileSystem, str)), \
            "Error: fileSystem argument must be a string."

        assert (fileSystem == "hadoop") or (fileSystem == "local"), \
            "Error, fileSystem argument only can be 'local' or 'hadoop'"


        if fileSystem == "hadoop":
            folderPath = path + "/" + "checkPointFolder"
            self.deleteCheckPointFolder(path=path, fileSystem=fileSystem)

            # Creating file:
            print ("Creation of hadoop folder...")
            command = "hadoop fs -mkdir " + folderPath
            print ("$" + command)
            os.system(command)
            print ("Hadoop folder created. \n")

            print ("Setting created folder as checkpoint folder...")
            self.__sc.setCheckpointDir(folderPath)
            print ("Done.")


        else:
            # Folder path:
            folderPath = path + "/" + "checkPointFolder"
            # Checking if tempFolder exits:
            print ("Deleting previous folder if exists...")
            if os.path.isdir(folderPath):

                # Deletes folder if exits:
                rmtree(folderPath)
                print ("Creation of checkpoint directory...")
                # Creates new folder:
                os.mkdir(folderPath)
                print ("Done.")
            else:
                print ("Creation of checkpoint directory...")

                # Creates new folder:
                os.mkdir(folderPath)
                print ("Listo.")

            self.__sc.setCheckpointDir(dirName = "file://" + folderPath)

    @classmethod
    def deleteCheckPointFolder(self, path, fileSystem):
        """Function that deletes the temporal folder where temp files were stored.
        The path required is the same provided by user in setCheckPointFolder().

        :param fileSystem   Describes if file system is local or hadoop file system.
        """
        assert (isinstance(fileSystem, str)), "Error: fileSystem argument must be a string."

        assert (fileSystem == "hadoop") or (fileSystem == "local"), \
            "Error, fileSystem argument only can be 'local' or 'hadoop'"

        if fileSystem == "hadoop":
            # Folder path:
            folderPath = path + "/" + "checkPointFolder"
            print ("Deleting checkpoint folder...")
            command = "hadoop fs -rm -r " + folderPath
            os.system(command)
            print ("$" + command)
            print ("Folder deleted. \n")

        else:
            print ("Deleting checkpoint folder...")
            # Folder path:
            folderPath = path + "/" + "checkPointFolder"
            # Checking if tempFolder exits:
            if os.path.isdir(folderPath):
                # Deletes folder if exits:
                rmtree(folderPath)
                # Creates new folder:
                print ("Folder deleted. \n")
            else:
                print ("Folder deleted. \n")
                pass

    @classmethod
    def getColumnNamesbyType(self, df, dataType):
        """This function returns column names of dataFrame which have the same
        datatype provided. It analyses column datatype by dataFrame.dtypes method.

        :return    List of column names of a type specified.
        """
        assert (dataType in ['string', 'integer', 'float', 'date', 'double']), \
            "Error, dataType only can be one of the following values: 'string', 'integer', 'float', 'date', 'double'"

        dicc = {'string': 'string','integer': 'int','float': 'float', 'double': 'double'}

        return list(y[0] for y in filter(lambda x: x[1] == dicc[dataType], df.dtypes))



class Airtable():
    def __init__(self, path):
        # Setting airtable dataset variable
        self.__airtable = 0
        # Setting SQLContext as a global variable of the class
        self.spark = SparkSession()
        self.sc = self.spark.sparkContext
        # Reading dataset
        self.readAirtableCsv(path)

    def readAirtableCsv(self, path):
        """This function reads the airtable or csv file that user have to fill when the original dataframe
        is analyzed.

        The path could be a dowload link or a path from a local directory.
        When a local directory is provided, it is necessary to write the corresponding prefix. i.e.:
        file://.... or hdfs://...
        """

        if re.match("https", path):
            print("Reading file from web...")
            # Doing a request to url:
            response = urllib.request.urlopen(path)
            # Read obtained data:
            data = response.read()      # a `bytes` object
            # Decoding data:
            text = data.decode('utf-8')
            # Here text is splitted by ',' adn replace
            values =  [line.replace("\"", "").split(",") for line in text.split("\n")]

            # Airtable datafra that corresponds to airtable table:
            self.__airtable = self.sc.parallelize(values[1:]).toDF(values[0])
            print("Done")

        else:
            print("Reading local file...")
            self.__airtable = self.spark.read \
            .format('csv') \
            .options(header='true') \
            .options(delimiter=",") \
            .options(inferSchema='true') \
            .load(path)
            print("Done")


    def setAirtableDF(self, df):
        """This function set a dataframe into the class for subsequent actions.
        """
        assert isinstance(df, pyspark.sql.dataframe.DataFrame), "Error: df argument must a sql.dataframe type"
        self.__airtable = df

    def getAirtableDF(self):
        """This function return the dataframe of the class"""
        return self.__airtable

    @classmethod
    def extractColFromDF(self, df, column):
        """This function extract column from dataFrame.
        Returns a list of elements of column specified.

        :param df   Input dataframe
        :param column Column to extract from dataframe provided.
        :return list of elements of column extracted
        """
        return [x[0] for x in df.select(column).collect()]

    def readColNames(self):
        """This function extract the column names from airtable dataset.

        :return list of tuples: [(oldNameCol1, newNameCol1),
                (oldNameCol2, newNameCol2)...]

                Where oldNameColX are the original names of columns in
                dataset, names are extracted from the airtable
                dataframe column 'Feature Name'.

                On the other hand, newNameColX are the new names of columns
                of dataset. New names are extracted from the airtable
                dataframe  'Code Name'
                                 """

        return list(zip(self.extractColFromDF(self.__airtable, 'Feature Name'),
                        self.extractColFromDF(self.__airtable, 'Code Name')))

    def readNewColNames(self):
        """This function extract the new column names from airtable dataset.

        :return list of new column names. Those names are from column 'Code Name'
                                 """

        return self.extractColFromDF(self.__airtable, 'Code Name')

    def readOldColNames(self):
        """This function extract the old column names from airtable dataset.

        :return list of old column names. Those names are from column 'Feature Name'
        """
        return self.extractColFromDF(self.__airtable, 'Feature Name')

    def readColTypes(self, names='newNames'):
        """This function extract the types of columns detailled in airtable dataset.

        :param  names   Name columns of dataFrame.  names variable consist of a list
                of columns read it from airtable dataFrame in the column specified.
                 names as argument can only have two values as input of this function:
                 'newNames' and 'oldNames. When 'newNames' is provided, the names are
                 read from 'Code Name' airtable dataFrame. On the other hand, when
                 'oldNames' is provided, 'Feature Name' is read it from airtable
                 dataFrame.


                 Column names are read using the function readNewColNames(self).
                 Types are extracted from 'DataType'

        :return list of colNames and their data types in the following form:
                [(colName1, 'string'), (colName2, 'integer'), (colName3, 'float')]
        """

        assert (names == 'newNames') or (names == 'oldNames'), "Error, names argument" \
                                                                "only can be the following values:" \
                                                                "'newNames' or 'oldNames'"

        colNames = {'newNames': 'Code Name', 'oldNames': 'Feature Name'}
        return list(zip(self.extractColFromDF(self.__airtable, colNames[names]),
                        self.extractColFromDF(self.__airtable, 'DataType')))




import findspark
findspark.init()

import logging
import pytest

from pyspark.sql.session import SparkSession
from pyspark import SparkConf
from pyspark import SparkContext
from pyspark.streaming import StreamingContext


def quiet_py4j():
    """ turn down spark logging for the test context """
    logger = logging.getLogger('py4j')
    logger.setLevel(logging.WARN)


@pytest.fixture(scope="session")
def spark_context(request):
    """ fixture for creating a spark context
    Args:
        request: pytest.FixtureRequest object
    """
    conf = (SparkConf().setMaster("local[2]").setAppName("pytest-pyspark-local-testing"))
    sc = SparkContext(conf=conf)
    request.addfinalizer(lambda: sc.stop())

    quiet_py4j()
    return sc


@pytest.fixture(scope="session")
def hive_context(spark_context):
    """  fixture for creating a Hive Context. Creating a fixture enables it to be reused across all
        tests in a session
    Args:
        spark_context: spark_context fixture
    Returns:
        HiveContext for tests
    """
    return SparkSession.builder.getOrCreate()


@pytest.fixture(scope="session")
def streaming_context(spark_context):
    return StreamingContext(spark_context, 1)



# -*- coding: utf-8 -*-
# Importing sql functions
from pyspark.sql.functions import udf, when, col, min as cmin, max as cmax
from pyspark.sql.types import StringType, DoubleType
# Importing plotting libraries
import matplotlib.pyplot as plt
# Importing numeric module
import numpy as np
# Importing features to build tables
from IPython.display import display
# Importing time librarie to measure time
import time
# Importing dumps
from json import dumps
import pyspark.sql.dataframe

#Import profiler
import spark_df_profiling_optimus

class ColumnTables():
    """This class builds a table to describe the number of the different dataTypes in a column dataFrame.

    It is important to notice that this is not the best way to build a table. It will be better if a general
    building table class is built"""
    def __init__(self, colName, dataTypeInferred, qtys, percents):
        self.qtys = qtys
        self.percents = percents
        self.colName = colName
        self.dataTypeInferred = dataTypeInferred
        self.labelsTable = ["None", "Empty str", "String", "Integer", "Float"]

    def _repr_html_(self):
        # Creation of blank table:
        html = ["<table width=50%>"]

        # Adding a row:
        html.append("<tr>")
        html.append("<td colspan=3 >{0}</td>".format("<b> Column name: </b>" + self.colName))
        html.append("</tr>")

        # Adding a row:
        html.append("<tr>")
        html.append("<td colspan=3 >{0}</td>".format("<b> Column datatype: </b>" + self.dataTypeInferred))
        html.append("</tr>")

        # Adding a row:
        html.append("<tr>")
        html.append("<th>{0}</td>".format('Datatype'))
        html.append("<th>{0}</td>".format('Quantity'))
        html.append("<th>{0}</td>".format('Percentage'))
        html.append("</tr>")

        for ind in range(len(self.labelsTable)):
            html.append("<tr>")
            html.append("<td>{0}</td>".format(self.labelsTable[ind]))
            html.append("<td>{0}</td>".format(self.qtys[ind + 1]))
            html.append("<td>{0}</td>".format("%0.2f" % self.percents[ind] + " %"))
            html.append("</tr>")

        html.append("</table>")
        return ''.join(html)


# Input data:
class GeneralDescripTable():
    def __init__(self, fileName, columnNumber, rowNumber):
        self.labels = ['File Name', "Columns", "Rows"]
        self.dat = [fileName, columnNumber, rowNumber]

    def _repr_html_(self):
        # Creation of blank table:
        html = ["<table width=50%>"]
        # Adding a row:
        html.append("<tr>")
        html.append("<th colspan=3>{0}</td>".format('General description'))
        html.append("</tr>")

        # Adding a row:
        html.append("<tr>")
        html.append("<th colspan=1>{0}</td>".format('Features'))
        html.append("<th colspan=2>{0}</td>".format('Name or Quantity'))
        html.append("</tr>")

        for ind in range(len(self.dat)):
            # Adding a row:
            html.append("<tr>")
            html.append("<th colspan=1>{0}</td>".format(self.labels[ind]))
            html.append("<td colspan=2>{0}</td>".format(self.dat[ind]))
            html.append("</tr>")

        return ''.join(html)
        # self.body =  ''.join(html)


class DataTypeTable():
    def __init__(self, lista):
        self.lista = lista
        # self.tuples = tuples
        self.html = ""

    @classmethod
    def colTable(self, lista):
        html = []
        html.append("<table width=100%>")
        for x in lista:
            html.append("<tr >")
            html.append("<td style='text-align: center'>")

            html.append("<div style='min-height: 20px;'>")
            html.append(str(x))
            html.append("</div>")
            html.append("</td>")
            html.append("</tr>")
        html.append("</table>")
        return "".join(html)

    def _repr_html_(self):
        self.html = ["<table width=50%>"]
        headers = ['integers', 'floats', 'strings']

        # This verify is some of list of types is empty.
        # The idea is not to draw an empty table.
        ver = [x != [] for x in self.lista]

        # First row of table:
        self.html.append("<tr>")
        for x in range(len(headers)):
            if ver[x]:  # If list is not empty, print its head
                self.html.append("<th style='text-align: center'>")
                self.html.append(str(headers[x]))
                self.html.append("</th>")
        self.html.append("</tr>")

        # Adding the second row:
        self.html.append("<tr>")
        for x in range(len(self.lista)):
            if ver[x]:
                # Adding a td
                self.html.append("<td style='vertical-align: top;text-align: center;'>")
                # Adding a table:


                self.html.append(self.colTable(self.lista[x]))

                self.html.append("</td>")
        self.html.append("</tr>")

        self.html.append("</table>")
        return ''.join(self.html)

# This class makes a profile for a given dataframe and its different general features.
# Based on spark-df-profiling
class DataFrameProfiler():

    def __init__(self, df):
        # Asserting if df is dataFrame datatype.
        assert (isinstance(df, pyspark.sql.dataframe.DataFrame)), \
            "Error, df argument must be a pyspark.sql.dataframe.DataFrame instance"

       # Dataframe
        self.__df = df
        self.__df.cache()

    def profiler(self):
        """
        This function calls the ProfileReport method from spark-df-profiling-optimus,
        it gets the current DF in the analyzer and them returns the HTML profile"
        :return: Profile of the DF in HTML format embedded in the Notebook
        """
        dfProfiler = self.__df
        return spark_df_profiling_optimus.ProfileReport(dfProfiler)


# This class makes an analisis of dataframe datatypes and its different general features.
class DataFrameAnalizer():
    def __init__(self, df, pathFile, pu=0.1, seed=13):
        # Asserting if df is dataFrame datatype.
        assert (isinstance(df, pyspark.sql.dataframe.DataFrame)), \
            "Error, df argument must be a pyspark.sql.dataframe.DataFrame instance"
        # Asserting if path specified is string datatype
        assert (isinstance(pathFile, str)), \
            "Error, pathFile argument must be string datatype."
        # Asserting if path includes the type of filesystem
        assert (("file:///" == pathFile[0:8]) or ("hdfs:///" == pathFile[0:8])), \
            "Error: path must be with a 'file://' prefix \
            if the file is in the local disk or a 'path://' \
            prefix if the file is in the Hadood file system"

        # Asserting if seed is integer
        assert isinstance(seed, int), "Error, seed must be an integer"

        # Asserting pu is between 0 and 1.
        assert (pu <= 1) and (pu >= 0), "Error, pu argument must be between 0 and 1"

        # Dataframe
        self.__rowNumber = 0
        self.__pathFile = pathFile
        self.__currentlyObserved = 0  # 0 => whole 1 => partition
        self.__df = df
        self.__df.cache()
        self.__sample = df.sample(False, pu, seed)

    @classmethod
    def __createDict(self, keys, values):
        """This functions is a helper to build dictionaries. The first argument must be a list of keys but it
        can be a string also (in this case the string will be packaged into a list. The keys provided
        will be the dictionary keys. The second argument represents the values of each key. values argument can
        be a list or another dictionary."""

        dicc = {}
        # Assert if keys is a string, if not, the string is placed it inside a list
        if isinstance(keys, str):
            keys = [keys]
        # If values is not a list, place it inside a list also
        if not isinstance(values, list):
            values = [values]
        # Making dictionary
        for index in range(len(keys)):
            dicc[keys[index]] = values[index]
        # Return dictionary built
        return dicc

    @classmethod
    def __verification(self, tempDf, columnName):
        # Function for determine if register value is float or int or string:
        def dataType(value):
            if isinstance(value, int):  # Check if value is integer
                return 'integer'
            elif isinstance(value, float):
                return 'float'
            elif isinstance(value, str):
                try:  # Try to parse (to int) register value
                    int(value)
                    # Add 1 if suceed:
                    return 'integer'
                except ValueError:
                    try:
                        # Try to parse (to float) register value
                        float(value)
                        # Add 1 if suceed:
                        return 'float'
                    except ValueError:
                        # Then, it is a string
                        return 'string'
            else:
                return 'null'

        func = udf(dataType, StringType())
        tempDf = tempDf.withColumn('types', func(col(columnName)))

        types = tempDf.groupBy('types').count()

        typeslabels = ['integer', 'float', 'string', 'null']

        numbers = {}
        for row in types.collect(): numbers[row[0]] = row[1]

        # completing values of rest types:
        for label in typeslabels:
            if not label in numbers:
                numbers[label] = 0

        numberEmptyStrs = tempDf.where(col(columnName) == '').count()
        numbers['string'] = numbers['string'] - numberEmptyStrs

        # List of returning values:
        values = [tempDf, numbers['null'], numberEmptyStrs, numbers['string'], numbers['integer'], numbers['float']]
        return values

        # Analize of each column:
    def __analize(self, dfColAnalizer, column, rowNumber, plots, printType, valuesBar, numBars, typesDict):
        t = time.time()
        sampleTableDict = {'string': 0., 'integer': 0, 'float': 0}
        # Calling verification ruotine to obtain datatype's counts
        # returns: [dataframeColumn, number of nulls, number of strings, number of integers, number of floats]
        dtypeNumbers = self.__verification(dfColAnalizer.select(column), column)

        def validColCheck(f):
            """This functions analyze if column has more than two different data types
            and return True if the column analyzed is found to be with different datatypes"""
            # Verifying if there is a dummy column:
            dummy = any(x == 1 for x in f[0:2]) and all(x == 0 for x in f[2:])

            # Verifying if there is a column with different datatypes:
            # Sum the first and the second number:
            sumFirstAndSecond = lambda lista: [lista[x] + lista[0] if x == 1 else lista[x] for x in range(len(lista))][
                                              1:]
            # Check if the column has different datatypes:
            differentTypes = sum([1 if x == 0 else 0 for x in sumFirstAndSecond(f[1:])]) < 2

            return dummy or differentTypes

        # Calculate percentages of datatypes:
        percentages = [(dtypeNumber / float(rowNumber)) * 100 for dtypeNumber in dtypeNumbers[1:]]

        if (validColCheck(dtypeNumbers[1:])):
            invalidCols = column
        else:
            invalidCols = False

        # Instance of columnTables to display results:
        typeCol = dfColAnalizer.select(column).dtypes[0][1]

        tempObj = ColumnTables(column, typeCol, dtypeNumbers, percentages)
        display(tempObj)

        if typeCol != 'string':
            print("Min value: ", dfColAnalizer.select(cmin(col(column))).first()[0])
            print("Max value: ", dfColAnalizer.select(cmax(col(column))).first()[0])

        # Plot bar stack:
        # if plots==True: self.__bar_stack_type(percentages, column)

        if printType == True:
            typeslabels = ['integer', 'float', 'string']
            listOfEachtype = [dtypeNumbers[0].where(col('types') == tipo).where(col(column) != '') \
                                  .drop('types').select(column).distinct() \
                                  .limit(numBars).rdd.map(lambda x: x[column]).collect() for tipo in typeslabels]
            display(DataTypeTable(listOfEachtype))
            sampleTableDict = self.__createDict(typeslabels, listOfEachtype)

        # Obtaining number of strings and numbers to decide what type of histogram (numerical
        # or categorical is needed)
        stringQty = dtypeNumbers[3]
        numberQty = np.sum(dtypeNumbers[4:])

        # Plotting histograms:
        # If number of strings is greater than (number of integers + number of floats)
        if (stringQty > numberQty):
            # Building histogram:
            histDict = self.getCategoricalHist(dfColAnalizer.select(column), numBars)
            histPlot = {"data": histDict}

            if plots == True:
                self.plotHist(
                    dfOneCol=dfColAnalizer.select(column),
                    histDict=histDict,
                    typeHist='categorical',
                    numBars=numBars,
                    valuesBar=valuesBar)
                plt.show()


        elif (stringQty < numberQty):
            # Building histogram:
            histDict = self.getNumericalHist(dfColAnalizer.select(column), numBars)
            histPlot = {"data": histDict}

            if plots == True:
                # Create the general blog and the "subplots" i.e. the bars
                histPlot = self.plotHist(dfColAnalizer.select(column),
                                        histDict=histDict,
                                        typeHist='numerical',
                                        numBars=numBars,
                                        valuesBar=valuesBar)
                plt.show()

        else:
            print("No valid data to print histogram or plots argument set to False")

            histPlot = {"data": [{"count": 0, "value": "none"}]}




        numbers = list(dtypeNumbers[1:])
        valid_values = self.__createDict(['total', 'string', 'integer', 'float'],
                                         [int(np.sum(dtypeNumbers[-3:])),
                                          numbers[-3],
                                          numbers[-2],
                                          numbers[-1]]
                                         )

        missing_values = self.__createDict(
            ['total', 'empty', 'null'],
            [int(np.sum(numbers[0:2])), numbers[1], numbers[0]
             ])

        # returns: [number of nulls, number of strings, number of integers, number of floats]
        summaryDict = self.__createDict(
            ["name", "type", "total", "valid_values", "missing_values"],
            [column, typesDict[typeCol], rowNumber, valid_values, missing_values
             ])

        columnDict = self.__createDict(
            ["summary", "graph", "sample"],
            [summaryDict, histPlot, sampleTableDict]
        )

        print ("end of __analyze", time.time() - t)
        return invalidCols, percentages, numbers, columnDict


    # This function, place values of frequency in histogram bars.
    @classmethod
    def __valuesOnBar(self, plotFig):
        rects = plotFig.patches
        for rect in rects:
            # Getting height of bars:
            height = rect.get_height()
            # Plotting texts on bars:
            plt.text(rect.get_x() + rect.get_width() / 2.,
                     1.001 * height, "%.2e" % int(height),
                     va='bottom', rotation=90)

    def __plotNumHist(self, histDict, column, valuesBar):
        values = [list(lista) for lista in list(zip(*[(dic['value'], dic['cont']) for dic in histDict]))]

        bins = values[0]

        if len(bins) == 1:
            width = bins[0] * 0.3
            bins[0] = bins[0] - width
        else:
            width = min(abs(np.diff(bins))) * 0.15

        # Plotting histogram:
        plotFig = plt.bar(np.array(bins) - width, values[1], width=width)

        if valuesBar == True: self.__valuesOnBar(plotFig)

        if len(bins) == 1:
            plt.xticks(np.round(bins))
            plt.xlim([0, bins[0] + 1])
        else:
            plt.xticks(np.round(bins))

        plt.ylim([0, np.max(values[1]) * 1.25])
        # Plot Title:
        plt.title(column)
        # Limits Y axes
        plt.show()

    def __plotCatHist(self, histDict, column, valuesBar):
        # Extracting values from dictionary
        k = list(filter(lambda k: k != 'cont', histDict[0].keys()))[0]

        values = [list(lista) for lista in list(zip(*[(dic[k], dic['cont']) for dic in histDict]))]
        index = np.arange(len(values[0]))

        # Plot settings
        fig, ax = plt.subplots()
        # We need to draw the canvas, otherwise the labels won't be positioned and
        # won't have values yet.
        fig.canvas.draw()
        # Setting values of xticks
        ax.set_xticks(index)
        # Setting labels to the ticks
        ax.set_xticklabels(values[0], rotation=90)

        # Plot of bars:
        width = 0.5
        bins = index - width / 2
        plotFig = plt.bar(bins, values[1], width=width)

        # If user want to see values of frequencies over each bar
        if valuesBar == True: self.__valuesOnBar(plotFig)

        # Plot Title:
        plt.title(column)
        # Limits Y axes
        plt.ylim([0, np.max(values[1]) * 1.3])
        plt.xlim([-1, index[-1] + 1])
        plt.show()

    def __swapStatus(self):
        self.__exchangeData()  # exchange data
        self.__currentlyObserved = 0 if self.__currentlyObserved == 1 else 1

    def __exchangeData(self):  # Swaps the data among DF and the Sample
        if self.__currentlyObserved == 0:
            self.__hiddenData = self.__df
            self.__df = self.__sample

        else:
            self.__df = self.__hiddenData

    def analyzeSample(self):
        if self.__currentlyObserved == 0:
            self.__swapStatus()

    def analyzeCompleteData(self):
        if self.__currentlyObserved == 1:
            self.__swapStatus()

    def unpersistDF(self):
        self.__df.unpersist()

    def setDataframe(self, df):
        """This function set a dataframe into the class for subsequent actions.
        """
        # assert isinstance(df, pyspark.sql.dataframe.DataFrame), "Error: df argument must a sql.dataframe type"
        self.__df = df

    def getDataframe(self):
        """This function return the dataframe of the class"""
        return self.__df

    # Function to give general features of dataFrame:
    def generalDescription(self):
        # General data description:
        # Filename:
        fileName = self.__pathFile.split('/')[-1]
        # Counting the number of columns:
        rowNumber = self.__df.count()
        # Counting the number of rows:
        colNumber = len(self.__df.columns)
        # Writting General Description table:
        tempObj = GeneralDescripTable(fileName, colNumber, rowNumber)
        display(tempObj)

        dataTypes = len(set([x[1] for x in self.__df.dtypes]))

        return self.__createDict(["filename", "size", "columns", "rows", "datatypes"],
                                 [fileName, 500, colNumber, rowNumber, dataTypes])

    # Funtion to analize column datatypes, it also plot proportion datatypes and histograms:
    def columnAnalize(self, columnList, plots=True, valuesBar=True, printType=False, numBars=10):
        """
        # This function counts the number of registers in a column that are numbers (integers, floats) and the number of
        # string registers.
        # Input:
        # columnList: a list or a string column name
        # valuesBar (optional): Can be True or False. If it is True, frequency values are placed over each bar.
        # printType (optional): Can be True or False. If true it will print the analysis.
        # numBars: number of bars printed in histogram
        # Output:
        # values: a list containing the number of the different datatypes [nulls, strings, integers, floats]
        """
        # Asserting data variable columnList:
        assert isinstance(columnList, (str, list)), "Error: columnList has to be a string or list."

        # Asserting if valuesBar is type Boolean
        assert isinstance(valuesBar, bool), "Error: valuesBar must be boolean, True or False."

        # Asserting if printType is type Boolean
        assert isinstance(printType, bool), "Error: printType must be boolean. True or False."

        # Counting
        time1 = time.time()

        dfColAnalizer = self.__df

        # Column assignation:
        columns = []

        # If columnList is a string, convert it in a list:
        if isinstance(columnList, str):
            if columnList == "*":
                columns = dfColAnalizer.columns
            else:
                columns = [columnList]

        else:
            if len(columnList) > 1:
                columns = columnList

        # Asserting if columns provided are in dataFrame:
        assert all(
            col in dfColAnalizer.columns for col in columns), 'Error: Columns or column does not exist in the dataFrame'

        types = {"string": "ABC", "int": "#", "integer": "#", "float": "##.#", "double": "##.#", "bigint": "#"}

        rowNumber = self.__df.count()

        invalidCols = []
        jsonCols = []
        # In case first is not set, this will analize all columns
        for column in columns:
            # Calling function analize:
            invCol, percentages, dtypeNumbers, d = self.__analize(
                dfColAnalizer.select(column),
                column,
                rowNumber,
                plots,
                printType,
                valuesBar=valuesBar,
                numBars=numBars,
                typesDict=types)

            # Save the invalid col if exists
            invalidCols.append(invCol)

            jsonCols.append(d)

        time2 = time.time()
        print("Total execution time: ", (time2 - time1))

        invalidCols = list(filter(lambda x: x != False, invalidCols))

        jsonCols = self.__createDict(["summary", "columns"], [self.generalDescription(), jsonCols])
        return invalidCols, jsonCols

    def plotHist(self, dfOneCol, histDict, typeHist, numBars=20, valuesBar=True):
        """
        This function builds the histogram (bins) of an categorical column dataframe.
        Inputs:
        dfOneCol: A dataFrame of one column.
        histDict: Python dictionary with histogram values
        typeHist: type of histogram to be generated, numerical or categorical
        numBars: Number of bars in histogram.
        valuesBar: If valuesBar is True, values of frequency are plotted over bars.
        Outputs: dictionary of the histogram generated
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        Example:
        self.plotHist(df[column], typeHist='categorical', valuesBar=True)
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        """

        assert isinstance(typeHist, str), "Error, typeHist argument provided must be a string."
        assert typeHist == ('categorical') or (
            typeHist == 'numerical'), "Error, typeHist only can be 'categorical' or 'numerical'."

        # Getting column of dataframe provided
        column = dfOneCol.columns[0]

        if typeHist == 'categorical':
            # Plotting histogram
            self.__plotCatHist(histDict, column, valuesBar=True)
        else:
            # Plotting histogram
            self.__plotNumHist(histDict, column, valuesBar=True)

    def getCategoricalHist(self, dfOneCol, numBars):
        """This function analyzes a dataframe of a single column (only string type columns) and
        returns a dictionary with bins and values of frequency.

        :param dfOneCol     One column dataFrame.
        :param numBars      Number of bars or histogram bins.
        :return histDict    Python dictionary with histogram values and bins."""

        assert isinstance(numBars, int), "Error, numBars argument must be a string dataType."
        assert len(dfOneCol.columns) == 1, "Error, Dataframe provided must have only one column."
        assert dfOneCol.dtypes[0][1] == 'string', "Error, Dataframe column must be string data type."

        # Column Name
        column = dfOneCol.columns[0]

        dfKeys = dfOneCol.select(column) \
            .groupBy(column).count() \
            .sort('count', ascending=False) \
            .limit(numBars)

        # Extracting information dataframe into a python dictionary:
        histDict = []
        for row in dfKeys.collect():
            histDict.append(self.__createDict(['value', 'cont'], [row[0], row[1]]))

        return histDict

    def getNumericalHist(self, dfOneCol, numBars):
        """This function analyzes a dataframe of a single column (only numerical columns) and
        returns a dictionary with bins and values of frequency."""

        assert len(dfOneCol.columns) == 1, "Error, Dataframe provided must have only one column."

        # Column Name
        column = dfOneCol.columns[0]

        tempDf = dfOneCol.withColumn(column, col(column).cast('float')).na.drop(subset=column)

        # If we obtain a null column:
        assert not isinstance(tempDf.first(), type(None)), \
            "Error, Make sure column dataframe has numerical features. One of the first actions \
        getNumericalHist function does is a convertion dataType from original datatype \
        to float. If the column provided has only values that are \
        not numbers parseables to float, it will flag this error."

        # Getting min and max values:
        minValue = tempDf.select(cmin(col(column))).first()[0]
        maxValue = tempDf.select(cmax(col(column))).first()[0]

        # Numero de valores entre el máximo y el minimo:
        stepsValue = (maxValue - minValue) / (numBars)

        # if stepsValue is different to zero, for example, there is only one number distinct in columnName
        if stepsValue == 0:
            binsValues = [0, maxValue]  # Only one bin is generated
        else:
            # Intervals between min an max values of columnName are made based in number of numBars wanted
            binsValues = np.arange(minValue, maxValue + stepsValue, stepsValue)

        # Valores unicos:
        uniValues = [row[0] for row in tempDf.select(column).distinct().take(numBars * 100)]

        # Si la cantidad de bins es menor que los valores unicos, entonces se toman los valores unicos como bin.
        if len(binsValues) < len(uniValues):

        # This function search over columnName dataFrame to which interval belongs each cell
        # It returns the columnName dataFrame with an additional columnName which describes intervals of each columnName cell.
            def generateExpr(columnName, listIntervals):
                if (len(listIntervals) == 1):
                    return when(col(columnName).between(listIntervals[0][0], listIntervals[0][1]), 0).otherwise(None)
                else:
                    return (when((col(columnName) >= listIntervals[0][0]) & (col(columnName) < listIntervals[0][1]),
                             len(listIntervals) - 1)
                            .otherwise(generateExpr(columnName, listIntervals[1:])))

                # +--------+--------------------+
                # |columns |Number of list pairs|
                # +--------+--------------------+
                # |       5|                   4|
                # |       3|                   7|
                # |       6|                   3|
                # |       9|                   0|
                # |       1|                   9|
                # |       6|                   3|
                # |       4|                   6|
                # +--------+--------------------+

        # Getting ranges from list: i.e. [(0,1),(1,2),(2,3),(3,4)]
        funcPairs = lambda liste: [(liste[x], liste[x + 1]) for x in range(0, len(liste) - 1)]
        ranges = funcPairs(binsValues)

        # Identifying to which group belongs each cell of column Dataframe and count them in order to get frequencies
        # for each searh interval.
        freqDf = tempDf.select(col(column), generateExpr(column, ranges).alias("value")) \
            .groupBy("value").count()

        # +-----------+-----+
        # |intervals  |count|
        # +-----------+-----+
        # |          0|    2|
        # |          1|    3|
        # |          4|    2|
        # |          5|    1|
        # |          6|    1|
        # |         10|    1|
        # +-----------+-----+

        # Reverting the order of the list ranges, because 0 in the last interval in list provided to get FreqDF
        # so it is more intuitive if the list of ranges is reverted. Then the first and second pair interval in ranges1
        # correspond to 0 and 1 interval in list
        ranges1 = list(reversed(ranges))

        # From intervals, bins are calculated as the average of min and max interval.
        bins = [np.mean([rmin, rmax]) for (rmin, rmax) in ranges1]

        func = udf(lambda x: float(bins[x]), DoubleType())

        # Setting position of bars according to bins and group intervals:
        freqDf = freqDf.na.drop().withColumn('value', func(col('value')))

        # Extracting information dataframe into a python dictionary:
        histDict = []
        for row in freqDf.collect():
            histDict.append(self.__createDict(['value', 'cont'], [row[0], row[1]]))

        return histDict

    def uniqueValuesCol(self, column):
        """This function counts the number of values that are unique and also the total number of values.
        Then, returns the values obtained.
        :param  column      Name of column dataFrame, this argument must be string type.
        :return         dictionary of values counted, as an example:
                        {'unique': 10, 'total': 15}
        """

        assert column, "Error, column name must be string type."

        total = self.__df.select(column).count()
        distincts = self.__df.select(column).distinct().count()

        return {'total': total, 'unique': distincts}

    @classmethod
    def writeJson(self, jsonCols, pathToJsonFile):

        # assert isinstance(jsonCols, dict), "Error: columnAnalyse must be run before writeJson function."

        jsonCols = dumps(jsonCols)

        with open(pathToJsonFile, 'w') as outfile:
            # outfile.write(str(jsonCols).replace("'", "\""))
            outfile.write(jsonCols)



import re

from setuptools import setup, find_packages

# Get version without importing, which avoids dependency issues
def get_version():
    with open('optimus/version.py') as version_file:
        return re.search(r"""__version__\s+=\s+(['"])(?P<version>.+?)\1""",
                         version_file.read()).group('version')
                         
import sys
if sys.version_info < (3, 5):
    raise RuntimeError('This version requires Python 3.5+')  # pragma: no cover

def readme():
    with open('Readme.txt') as f:
        return f.read()

install_requires = ['pytest', 'findspark','pytest-spark','spark-df-profiling-optimus','pyspark']
lint_requires = [
    'pep8',
    'pyflakes'
]

tests_require = ['pytest','mock', 'nose']

dependency_links = []
setup_requires=['pytest-runner']
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

setup(
    name='optimuspyspark',
    version=get_version(),
    author='Favio Vazquez',
    author_email='favio.vazquez@ironmussa.com',
    url='https://github.com/ironmussa/Optimus/',
    download_url = 'https://github.com/ironmussa/Optimus/archive/0.5.1.tar.gz',
    description=('Optimus is the missing library for cleaning and preprocessing data in a distributed fashion with pyspark.'),
    long_description=readme(),
    license='APACHE',
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require={
        'test': tests_require,
        'all': install_requires + tests_require,
        'docs': ['sphinx'] + tests_require,
        'lint': lint_requires
    },
    dependency_links=dependency_links,
    test_suite='nose.collector',
    include_package_data=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
    ],
    keywords = ['datacleaner','apachespark', 'spark', 'pyspark', 'data-wrangling'],
)



# Importing sql types
from pyspark.sql.types import StringType, IntegerType, FloatType, DoubleType, ArrayType
# Importing sql functions
from pyspark.sql.functions import col, udf, trim, lit, format_number, months_between, date_format, unix_timestamp, \
    current_date, abs as mag
from pyspark.ml.feature import MinMaxScaler, VectorAssembler
import re
import string
import unicodedata
import pyspark.sql.dataframe


class DataFrameTransformer():
    """DataFrameTransformer is a class to make transformations in dataFrames"""

    def __init__(self, df):
        """Class constructor.
        :param  df      DataFrame to be transformed.
        """
        assert (isinstance(df, pyspark.sql.dataframe.DataFrame)), \
            "Error, df argument must be a pyspark.sql.dataframe.DataFrame instance"

        # Dataframe
        self.__df = df
        # SparkContext:
        # self.__sqlContext = SQLContext(self.__df.sql_ctx)
        self.__sqlContext = self.__df.sql_ctx
        self.__numberOfTransformations = 0

    @classmethod
    def __assertTypeStrOrList(self, variable, nameArg):
        """This function asserts if variable is a string or a list dataType."""
        assert isinstance(variable, (str, list)), \
            "Error: %s argument must be a string or a list." % nameArg

    @classmethod
    def __assertTypeStr(self, variable, nameArg):
        """This function asserts if variable is a string or a list dataType."""
        assert isinstance(variable, str), \
            "Error: %s argument must be a string." % nameArg

    @classmethod
    def __assertColsInDF(self, columnsProvided, columnsDF):
        """This function asserts if columnsProvided exists in dataFrame.
        Inputs:
        columnsProvided: the list of columns to be process.
        columnsDF: list of columns's dataFrames
        """
        colNotValids = (set([column for column in columnsProvided]).difference(set([column for column in columnsDF])))
        assert (colNotValids == set()), 'Error: The following columns do not exits in dataFrame: %s' % colNotValids

    def __addTransformation(self):
        self.__numberOfTransformations += 1

        if (self.__numberOfTransformations > 50):
            self.checkPoint()
            self.__numberOfTransformations = 0

    def setDataframe(self, df):
        """This function set a dataframe into the class for subsequent actions.
        """
        assert isinstance(df, pyspark.sql.dataframe.DataFrame), "Error: df argument must a sql.dataframe type"
        self.__df = df

    def getDataframe(self):
        """This function return the dataframe of the class"""
        return self.__df

    def lowerCase(self, columns):
        """This function set all strings in columns of dataframe specified to lowercase.
        Columns argument must be a string or a list of string. In order to apply this function to all
        dataframe, columns must be equal to '*'"""

        func = lambda cell: cell.lower() if cell is not None else cell
        self.setCol(columns, func, 'string')
        return self

    def upperCase(self, columns):
        """This function set all strings in columns of dataframe specified to uppercase.
        Columns argument must be a string or a list of string. In order to apply this function to all
        dataframe, columns must be equal to '*'"""
        func = lambda cell: cell.upper() if cell is not None else cell
        self.setCol(columns, func, 'string')
        return self

    def checkPoint(self):
        """This method is a very useful function to break lineage of transformations. By default Spark uses the lazy
        evaluation approach in processing data: transformation functions are not computed into an action is called.
        Sometimes when transformations are numerous, the computations are very extensive because the high number of
        operations that spark needs to run in order to get the results.

        Other important thing is that apache spark usually save task but not result of dataFrame, so tasks are
        accumulated and the same situation happens.

        The problem can be deal it with the checkPoint method. This method save the resulting dataFrame in disk, so
         the lineage is cut.
        """

        # Checkpointing of dataFrame. One question can be thought. Why not use cache() or persist() instead of
        # checkpoint. This is because cache() and persis() apparently do not break the lineage of operations,
        print ("Saving changes at disk by checkpoint...")
        self.__df.rdd.checkpoint()
        self.__df.rdd.count()
        self.__df = self.__sqlContext.createDataFrame(self.__df.rdd, self.__df.schema)
        print ("Done.")

    execute = checkPoint

    def trimCol(self, columns):
        """This methods cut left and right extra spaces in column strings provided by user.
        :param columns   list of column names of dataFrame.
                If a string "*" is provided, the method will do the trimming operation in whole dataFrame.

        :return transformer object
        """

        # Function to trim spaces in columns with strings datatype
        def colTrim(columns):
            exprs = [trim(col(c)).alias(c)
                     if (c in columns) and (c in validCols)
                     else c
                     for (c, t) in self.__df.dtypes]
            self.__df = self.__df.select(*exprs)

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols

        # Columns
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        # Trimming spaces in columns:
        colTrim(columns)

        self.__addTransformation()

        # Returning the transformer object for able chaining operations
        return self

    def dropCol(self, columns):
        """This method eliminate the list of columns provided by user.
        :param columns      list of columns names or a string (a column name).

        :return transformer object
        """

        def colDrop(columns):
            exprs = filter(lambda c: c not in columns, self.__df.columns)
            self.__df = self.__df.select(*exprs)

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Columns
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        # Calling colDrop function
        colDrop(columns)

        self.__addTransformation()

        # Returning the transformer object for able chaining operations
        return self

    def replaceCol(self, search, changeTo, columns):
        """This method search the 'search' value in DataFrame columns specified in 'columns' in order to replace it
        for 'changeTo' value.


        :param search       value to search in dataFrame.
        :param changeTo     value used to replace the old one in dataFrame.
        :param columns      list of string column names or a string (column name). If columns = '*' is provided,
                            searching and replacing action is made in all columns of DataFrame that have same
                            dataType of search and changeTo.

        search and changeTo arguments are expected to be numbers and same dataType ('integer', 'string', etc) each other.
        olumns argument is expected to be a string or list of string column names.

        :return transformer object
        """

        def colReplace(columns):
            self.__df = self.__df.replace(search, changeTo, subset=columns)

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Asserting search parameter is a string or a number
        assert isinstance(search, str) or isinstance(search, float) or isinstance(search,
                                                                                  int), "Error: Search parameter must be a number or string"

        # Asserting changeTo parameter is a string or a number
        assert isinstance(changeTo, str) or isinstance(changeTo, float) or isinstance(changeTo,
                                                                                      int), "Error: changeTo parameter must be a number or string"

        # Asserting search and changeTo have same type
        assert isinstance(search, type(changeTo)), 'Error: Search and ChangeTo must have same datatype: Integer, String, Float'

        # Change
        types = {type(''): 'string', type(int(1)): 'int', type(float(1.2)): 'float', type(1.2): 'double'};

        validCols = [c for (c, t) in filter(lambda t: t[1] == types[type(search)], self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        colNotValids = (set([column for column in columns]).difference(set([column for column in validCols])))

        assert (
            colNotValids == set()), 'Error: The following columns do not have same datatype argument provided: %s' % colNotValids

        colReplace(columns)

        self.__addTransformation()

        # Returning the transformer object for able chaining operations
        return self

    def deleteRow(self, func):
        """This function is an alias of filter and where spark functions.
        :param func     func must be an expression with the following form:

                func = col('colName') > value.

                func is an expression where col is a pyspark.sql.function.
        """
        self.__df = self.__df.filter(func)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    def setCol(self, columns, func, dataType):
        """This method can be used to make math operations or string manipulations in row of dataFrame columns.

        :param columns      list of columns (or a single column) of dataFrame.
        :param func         function or string type which describe the dataType that func function should return.
        :param dataType     string indicating one of the following options: 'integer', 'string', 'double','float'.

        'columns' argument is expected to be a string or a list of columns names.
        It is a requirement for this method that the dataType provided must be the same to dataType of columns.
        On the other hand, if user writes columns == '*' the method makes operations in func if only if columns
        have same dataType that dataType argument.

        :return transformer object
        """
        dictTypes = {'string': StringType(), 'str': StringType(), 'integer': IntegerType(),
                     'int': IntegerType(), 'float': FloatType(), 'double': DoubleType(), 'Double': DoubleType()}

        Types = {'string': 'string', 'str': 'string', 'String': 'string', 'integer': 'int',
                 'int': 'int', 'float': 'float', 'double': 'double', 'Double': 'double'}

        try:
            function = udf(func, dictTypes[dataType])
        except KeyError:
            assert False, "Error, dataType not recognized"

        def colSet(columns, function):
            exprs = [function(col(c)).alias(c) if c in columns else c for (c, t) in self.__df.dtypes]
            try:
                self.__df = self.__df.select(*exprs)
            except Exception as e:
                print(e)
                assert False, "Error: Make sure if operation is compatible with row datatype."

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == Types[dataType], self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        colNotValids = (set([column for column in columns]).difference(set([column for column in validCols])))

        assert (
            colNotValids == set()), 'Error: The following columns do not have same datatype argument provided: %s' % colNotValids

        colSet(columns, function)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    # Drop
    def keepCol(self, columns):
        """This method keep only columns specified by user with columns argument in DataFrame.
        :param columns list of columns or a string (column name).

        :return transformer object
        """

        def colKeep(columns):
            exprs = filter(lambda c: c in columns, self.__df.columns)
            self.__df = self.__df.select(*exprs)

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Check is column if a string.
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        # Calling colDrop function
        colKeep(columns)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    def clearAccents(self, columns):
        """This function deletes accents in strings column dataFrames, it does not eliminate main characters,
        but only deletes special tildes.

        :param columns  String or a list of column names.

        """

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if isinstance(columns, str): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        colNotValids = (set([column for column in columns]).difference(set([column for column in validCols])))

        assert (
            colNotValids == set()), 'Error: The following columns do not have same datatype argument provided: %s' % colNotValids

        # Receives  a string as an argument
        def remove_accents(inputStr):
            # first, normalize strings:
            nfkdStr = unicodedata.normalize('NFKD', inputStr)
            # Keep chars that has no other char combined (i.e. accents chars)
            withOutAccents = u"".join([c for c in nfkdStr if not unicodedata.combining(c)])
            return withOutAccents

        function = udf(lambda x: remove_accents(x) if x != None else x, StringType())
        exprs = [function(col(c)).alias(c) if (c in columns) and (c in validCols) else c for c in self.__df.columns]
        self.__df = self.__df.select(*exprs)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    def removeSpecialChars(self, columns):
        """This function remove special chars in string columns, such as: .!"#$%&/()
        :param columns      list of names columns to be processed.

        columns argument can be a string or a list of strings."""

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if isinstance(columns, str):
            columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        colNotValids = (set([column for column in columns]).difference(set([column for column in validCols])))

        assert (
            colNotValids == set()), 'Error: The following columns do not have same datatype argument provided: %s' % colNotValids

        def rmSpecChars(inputStr):
            # Remove all punctuation and control characters
            for punct in (set(inputStr) & set(string.punctuation)):
                inputStr = inputStr.replace(punct, "")
            return inputStr

        # User define function that does operation in cells
        function = udf(lambda cell: rmSpecChars(cell) if cell != None else cell, StringType())

        exprs = [function(c).alias(c) if (c in columns) and (c in validCols)  else c for c in self.__df.columns]

        self.__df = self.__df.select(*exprs)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    def removeSpecialCharsRegex(self, columns,regex):
        """This function remove special chars in string columns using a regex, such as: .!"#$%&/()
        :param columns      list of names columns to be processed.
        :param regex        string that contains the regular expression

        columns argument can be a string or a list of strings."""

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if isinstance(columns, str):
            columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        colNotValids = (set([column for column in columns]).difference(set([column for column in validCols])))

        assert (
            colNotValids == set()), 'Error: The following columns do not have same datatype argument provided: %s' % colNotValids

        def rmSpecCharsRegex(inputStr,regex):
            for s in set(inputStr):
                inputStr = re.sub(regex, '', inputStr)
            return inputStr

        # User define function that does operation in cells
        function = udf(lambda cell: rmSpecCharsRegex(cell,regex) if cell != None else cell, StringType())

        exprs = [function(c).alias(c) if (c in columns) and (c in validCols)  else c for c in self.__df.columns]

        self.__df = self.__df.select(*exprs)

        self.__addTransformation()  # checkpoint in case

        # Returning the transformer object for able chaining operations
        return self

    def renameCol(self, columns):
        """"This functions change the name of columns datraFrame.
        :param columns      List of tuples. Each tuple has de following form: (oldColumnName, newColumnName).

        """
        # Asserting columns is string or list:
        assert isinstance(columns, list) and isinstance(columns[0], tuple), \
            "Error: Column argument must be a list of tuples"


        colNotValids = (
            set([column[0] for column in columns]).difference(set([column for column in self.__df.columns])))

        assert (colNotValids == set()), 'Error: The following columns do not exits in dataFrame: %s' % colNotValids

        oldNames = [column[0] for column in columns]

        notInType = filter(lambda c: c not in oldNames, self.__df.columns)

        exprs = [col(column[0]).alias(column[1]) for column in columns] + [col(column) for column in notInType]

        self.__addTransformation()  # checkpoint in case

        self.__df = self.__df.select(*exprs)
        return self

    def lookup(self, column, strToReplace, listStr=None):
        """This method search a list of strings specified in `listStr` argument among rows
        in column dataFrame and replace them for `StrToReplace`.

        :param  column      Column name, this variable must be string dataType.
        :param  strToReplace    string that going to replace all others present in listStr argument
        :param  listStr     List of strings to be replaced

        `lookup` can only be runned in StringType columns.


        """
        # Check if columns argument a string datatype:
        self.__assertTypeStr(column, "column")

        # Asserting columns is string or list:
        assert isinstance(strToReplace, (str, dict)), "Error: StrToReplace argument must be a string or a dict"

        if  isinstance(strToReplace,dict):
            assert (strToReplace != {}), "Error, StrToReplace must be a string or a non empty python dictionary"
            assert (
                listStr == None), "Error, If a python dictionary if specified, listStr argument must be None: listStr=None"

        # Asserting columns is string or list:
        assert isinstance(listStr, list) and listStr != [] or (
            listStr == None), "Error: Column argument must be a non empty list"

        if isinstance(strToReplace, str):
            assert listStr != None, "Error: listStr cannot be None if StrToReplace is a String, please you need to specify \
             the listStr string"

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        if isinstance(column, str):
            column = [column]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=column, columnsDF=self.__df.columns)

        # Asserting if selected column datatype and search and changeTo parameters are the same:
        colNotValids = (set(column).difference(set([column for column in validCols])))
        assert (colNotValids == set()), 'Error: The column provided is not a column string: %s' % colNotValids

        # User defined function to search cell value in list provide by user:
        if isinstance(strToReplace, str) and listStr is not None:

            def revisar(cell):
                if cell is not None and (cell in listStr):
                    return strToReplace
                else:
                    return cell

            func = udf(lambda cell: revisar(cell), StringType())
        else:
            def replaceFromDic(strTest):
                for key in strToReplace.keys():
                    if strTest in strToReplace[key]:
                        strTest = key
                return strTest

            func = udf(lambda cell: replaceFromDic(cell) if cell != None else cell, StringType())

        # Calling udf for each row of column provided by user. The rest of dataFrame is
        # maintained the same.
        exprs = [func(col(c)).alias(c) if c == column[0] else c for c in self.__df.columns]

        self.__df = self.__df.select(*exprs)

        self.__addTransformation()  # checkpoint in case

        return self

    def moveCol(self, column, refCol, position):
        """This funcion change column position in dataFrame.
        :param column   Name of the column to be moved in dataFrame. column argument must be a string.
        :param refCol   Name of reference column in dataFrame. This column will be a reference to place the
                        column to be moved.
        :param position Can be one of the following options: 'after' or 'before'. If 'after' is provided, column
                        provided will be placed just after the refCol selected."""
        # Columns of dataFrame
        columns = self.__df.columns

        # Check if columns argument a string datatype:
        self.__assertTypeStr(column, "column")

        # Check if column to be process are in dataframe
        self.__assertColsInDF(columnsProvided=[column], columnsDF=self.__df.columns)

        # Check if columns argument a string datatype:
        self.__assertTypeStr(refCol, "refCol")

        # Asserting parameters are not empty strings:
        assert (
            (column != '') and (refCol != '') and (position != '')), "Error: Input parameters can't be empty strings"

        # Check if refCol is in dataframe
        self.__assertColsInDF(columnsProvided=[refCol], columnsDF=self.__df.columns)

        # Check if columns argument a position string datatype:
        self.__assertTypeStr(position, "position")

        # Asserting if position is 'after' or 'before'
        assert (position == 'after') or (
            position == 'before'), "Error: Position parameter only can be 'after' or 'before'"

        # Finding position of column to move:
        findCol = lambda columns, column: [index for index, c in enumerate(columns) if c == column]
        newIndex = findCol(columns, refCol)
        oldIndex = findCol(columns, column)

        # if position is 'after':
        if position == 'after':
            # Check if the movement is from right to left:
            if newIndex[0] >= oldIndex[0]:
                columns.insert(newIndex[0], columns.pop(oldIndex[0]))  # insert and delete a element
            else:  # the movement is form left to right:
                columns.insert(newIndex[0] + 1, columns.pop(oldIndex[0]))
        else:  # If position if before:
            if newIndex[0] >= oldIndex[0]:  # Check if the movement if from right to left:
                columns.insert(newIndex[0] - 1, columns.pop(oldIndex[0]))
            elif newIndex[0] < oldIndex[0]:  # Check if the movement if from left to right:
                columns.insert(newIndex[0], columns.pop(oldIndex[0]))

        self.__df = self.__df[columns]

        self.__addTransformation()  # checkpoint in case

        return self

    def explodeTable(self, colId, col1, newColFeature, listToAssign):
        """
        This function can be used to split a feature with some extra information in order
        to make a new column feature.

        :param colId    column name of the columnId of dataFrame
        :param col1     column name of the column to be split.
        :param newColFeature        Name of the new column.
        :param listToAssign         List of values to be counted.

        Please, see documentation for more explanations about this method.

        """
        # Asserting if position is string or list:

        assert isinstance(listToAssign, list), "Error: listToAssign argument must be a list"

        # Asserting parameters are not empty strings:
        assert (
            (colId != '') and (col1 != '') and (newColFeature != '')), "Error: Input parameters can't be empty strings"

        # Check if col1 argument is string datatype:
        self.__assertTypeStr(col1, "col1")

        # Check if newColFeature argument is a string datatype:
        self.__assertTypeStr(newColFeature, "newColFeature")

        # Check if colId argument is a string datatype:
        self.__assertTypeStr(colId, "colId")

        # Check if colId to be process are in dataframe
        self.__assertColsInDF(columnsProvided=[colId], columnsDF=self.__df.columns)

        # Check if col1 to be process are in dataframe
        self.__assertColsInDF(columnsProvided=[col1], columnsDF=self.__df.columns)

        # subset, only PAQ and Tipo_Unidad:
        subdf = self.__df.select(colId, col1)

        # dataframe Filtered:
        dfMod = self.__df.where(self.__df[col1] != newColFeature)

        # subset de
        newColumn = subdf.where(subdf[col1] == newColFeature).groupBy(colId).count()

        # Left join:
        newColumn = newColumn.withColumnRenamed(colId, colId + '_other')

        for x in range(len(listToAssign)):
            if x == 0:
                exprs = (dfMod[colId] == newColumn[colId + '_other']) & (dfMod[col1] == listToAssign[x])
            else:
                exprs = exprs | (dfMod[colId] == newColumn[colId + '_other']) & (dfMod[col1] == listToAssign[x])

        dfMod = dfMod.join(newColumn, exprs, 'left_outer')

        # Cleaning dataframe:
        dfMod = dfMod.drop(colId + '_other').na.fill(0).withColumnRenamed('count', newColFeature)
        self.__df = dfMod

        self.__addTransformation()  # checkpoint in case

        return self

    def dateTransform(self, columns, currentFormat, outputFormat):
        """
        :param  columns     Name date columns to be transformed. Columns ha
        :param  currentFormat   currentFormat is the current string dat format of columns specified. Of course,
                                all columns specified must have the same format. Otherwise the function is going
                                to return tons of null values because the transformations in the columns with
                                different formats will fail.
        :param  outputFormat    output date string format to be expected.
        """
        # Check if currentFormat argument a string datatype:
        self.__assertTypeStr(currentFormat, "currentFormat")
        # Check if outputFormat argument a string datatype:
        self.__assertTypeStr(outputFormat, "outputFormat")
        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        if isinstance(columns, str):
            columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        exprs = [date_format(unix_timestamp(c, currentFormat).cast("timestamp"), outputFormat).alias(
            c) if c in columns else c for c in self.__df.columns]

        self.__df = self.__df.select(*exprs)

        self.__addTransformation()  # checkpoint in case

        return self

    def ageCalculate(self, column, dateFormat, nameColAge):
        """
        This method compute the age of clients based on their born dates.
        :param  column      Name of the column born dates column.
        :param  dateFormat  String format date of the column provided.
        :param  nameColAge  Name of the new column, the new columns is the resulting column of ages.

        """
        # Check if column argument a string datatype:
        self.__assertTypeStr(column, "column")

        # Check if dateFormat argument a string datatype:
        self.__assertTypeStr(dateFormat, "dateFormat")

        # Asserting if column if in dataFrame:
        assert column in self.__df.columns, "Error: Column assigned in column argument does not exist in dataFrame"

        # Output format date
        Format = "yyyy-MM-dd"  # Some SimpleDateFormat string

        exprs = format_number(
            mag(
                months_between(date_format(
                    unix_timestamp(column, dateFormat).cast("timestamp"), Format), current_date()) / 12), 4).alias(
            nameColAge)

        self.__df = self.__df.withColumn(nameColAge, exprs)

        self.__addTransformation()  # checkpoint in case

        return self

    def castFunc(self, colsAndTypes):
        """

        :param colsAndTypes     List of tuples of column names and types to be casted. This variable should have the
                                following structure:

                colsAndTypes = [('columnName1', 'integer'), ('columnName2', 'float'), ('columnName3', 'string')]

                The first parameter in each tuple is the column name, the second is the finale datatype of column after
                the transformation is made.
        :return:
        """

        dictTypes = {'string': StringType(), 'str': StringType(), 'integer': IntegerType(),
                     'int': IntegerType(), 'float': FloatType(), 'double': DoubleType(), 'Double': DoubleType()}

        Types = {'string': 'string', 'str': 'string', 'String': 'string', 'integer': 'int',
                 'int': 'int', 'float': 'float', 'double': 'double', 'Double': 'double'}

        # Asserting colsAndTypes is string or list:
        assert isinstance(colsAndTypes, (str, list)), "Error: Column argument must be a string or a list."

        if isinstance(colsAndTypes, str):
            colsAndTypes = [colsAndTypes]

        columnNames = [column[0] for column in colsAndTypes]

        # Check if columnNames to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columnNames, columnsDF=self.__df.columns)

        notSpecifiedColumns = filter(lambda c: c not in columnNames, self.__df.columns)

        exprs = [col(column[0]).cast(dictTypes[Types[column[1]]]).alias(column[0]) for column in colsAndTypes] + [
            col(column) for column in notSpecifiedColumns]

        self.__df = self.__df.select(*exprs)
        self.__addTransformation()  # checkpoint in case

        return self

    def emptyStrToStr(self, columns, customStr):
        """
        This function replace a string specified 
        """
        # Check if customStr argument a string datatype:
        self.__assertTypeStr(customStr, "customStr")

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Filters all string columns in dataFrame
        validCols = [c for (c, t) in filter(lambda t: t[1] == 'string', self.__df.dtypes)]

        # If None or [] is provided with column parameter:
        if (columns == "*"): columns = validCols[:]

        # If columns is string, make a list:
        if type(columns) == type(' '): columns = [columns]

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        def blank_as_null(x):
            return when(col(x) != "", col(x)).otherwise(customStr)

        exprs = [blank_as_null(c).alias(c) if (c in columns) and (c in validCols)  else c for c in self.__df.columns]

        self.__df = self.__df.select(*exprs)
        self.__addTransformation()  # checkpoint in case

        return self

    def operationInType(self, parameters):
        """ This function makes operations in a columnType of dataframe. It is well know that DataFrames are consistent,
        but it in this context, operation are based in types recognized by the dataframe analyzer, types are identified
        according if the value is parsable to int or float, etc.

        This functions makes the operation in column elements that are recognized as the same type that the dataType
        argument provided in the input function.

        Columns provided in list of tuples cannot be repeated
        :param parameters   List of columns in the following form: [(columnName, dataType, func),
                                                                    (columnName1, dataType1, func1)]
        :return None
        """

        def checkDataType(value):

            try:  # Try to parse (to int) register value
                int(value)
                # Add 1 if suceed:
                return 'integer'
            except ValueError:
                try:
                    # Try to parse (to float) register value
                    float(value)
                    # Add 1 if suceed:
                    return 'float'
                except ValueError:
                    # Then, it is a string
                    return 'string'
            except TypeError:
                return 'null'

        Types = {type('str'): 'string', type(1): 'int', type(1.0): 'float'}

        exprs = []
        for column, dataType, func in parameters:
            # Cheking if column name is string datatype:
            self.__assertTypeStr(column, "columnName")
            # Checking if column exists in dataframe:
            assert column in self.__df.columns, \
                "Error: Column %s specified as columnName argument does not exist in dataframe" % column
            # Checking if column has a valid datatype:
            assert (dataType in ['integer', 'float', 'string',
                                 'null']), \
                "Error: dataType only can be one of the followings options: integer, float, string, null."
            # Checking if func parameters is func dataType or None
            assert isinstance(func, type(None)) or isinstance(func, type(lambda x: x)), \
                "func argument must be a function or NoneType"

            if 'function' in str(type(func)):
                funcUdf = udf(lambda x: func(x) if checkDataType(x) == dataType else x)

            if isinstance(func, str) or isinstance(func, int) or isinstance(func, float):
                assert [x[1] in Types[type(func)] for x in filter(lambda x: x[0] == columnName, self.__df.dtypes)][
                    0], \
                    "Error: Column of operation and func argument must be the same global type. " \
                    "Check column type by df.printSchema()"
                funcUdf = udf(lambda x: func if checkDataType(x) == dataType else x)

            if func is None:
                funcUdf = udf(lambda x: None if checkDataType(x) == dataType else x)

            exprs.append(funcUdf(col(column)).alias(column))

        colNotProvided = [x for x in self.__df.columns if x not in [column[0] for column in parameters]]

        self.__df = self.__df.select(colNotProvided + [*exprs])
        self.__addTransformation()  # checkpoint in case

        return self

    def rowFilterByType(self, columnName, typeToDelete):
        """This function has built in order to deleted some type of dataframe """
        # Check if columnName argument a string datatype:
        self.__assertTypeStr(columnName, "columnName")
        # Asserting if columnName exits in dataframe:
        assert columnName in self.__df.columns, "Error: Column specified as columnName argument does not exist in dataframe"
        # Check if typeToDelete argument a string datatype:
        self.__assertTypeStr(typeToDelete, "typeToDelete")
        # Asserting if dataType argument has a valid type:
        assert (typeToDelete in ['integer', 'float', 'string',
                                 'null']), "Error: dataType only can be one of the followings options: integer, float, string, null."

        # Function for determine if register value is float or int or string:
        def dataType(value):

            try:  # Try to parse (to int) register value
                int(value)
                # Add 1 if suceed:
                return 'integer'
            except ValueError:
                try:
                    # Try to parse (to float) register value
                    float(value)
                    # Add 1 if suceed:
                    return 'float'
                except ValueError:
                    # Then, it is a string
                    return 'string'
            except TypeError:
                return 'null'

        func = udf(dataType, StringType())
        self.__df = self.__df.withColumn('types', func(col(columnName))).where((col('types') != typeToDelete)).drop(
            'types')
        self.__addTransformation()  # checkpoint in case

        return self

    def undoVecAssembler(self, column, featureNames):
        """This function unpack a column of list arrays into different columns.
        +-------------------+-------+
        |           features|columna|
        +-------------------+-------+
        |[11, 2, 1, 1, 1, 1]|   hola|
        | [0, 1, 1, 1, 1, 1]|  salut|
        |[31, 1, 1, 1, 1, 1]|  hello|
        +-------------------+-------+
                      |
                      |
                      V
        +-------+---+---+-----+----+----+---+
        |columna|one|two|three|four|five|six|
        +-------+---+---+-----+----+----+---+
        |   hola| 11|  2|    1|   1|   1|  1|
        |  salut|  0|  1|    1|   1|   1|  1|
        |  hello| 31|  1|    1|   1|   1|  1|
        +-------+---+---+-----+----+----+---+
        """
        # Check if column argument a string datatype:
        self.__assertTypeStr(column, "column")

        assert (column in self.__df.columns), "Error: column specified does not exist in dataFrame."

        assert (type(featureNames) == type([])), "Error: featureNames must be a list of strings."
        # Function to extract value from list column:
        func = udf(lambda x, index: x[index])

        exprs = []

        # Recursive function:
        def exprsFunc(column, exprs, featureNames, index):
            if index == 0:
                return [func(col(column), lit(index)).alias(featureNames[index])]
            else:
                return exprsFunc(column, exprs, featureNames, index - 1) + [
                    func(col(column), lit(index)).alias(featureNames[index])]

        self.__df = self.__df.select(
            [x for x in self.__df.columns] + [*exprsFunc(column, exprs, featureNames, len(featureNames) - 1)]).drop(
            column)
        self.__addTransformation()  # checkpoint in case

        return self

    def scaleVecCol(self, columns, nameOutputCol):
        """
        This function groups the columns specified and put them in a list array in one column, then a scale
        process is made. The scaling proccedure is spark scaling default (see the example
        bellow).

        +---------+----------+
        |Price    |AreaLiving|
        +---------+----------+
        |1261706.9|16        |
        |1263607.9|16        |
        |1109960.0|19        |
        |978277.0 |19        |
        |885000.0 |19        |
        +---------+----------+

                    |
                    |
                    |
                    V
        +----------------------------------------+
        |['Price', 'AreaLiving']                 |
        +----------------------------------------+
        |[0.1673858972637624,0.5]                |
        |[0.08966137157852398,0.3611111111111111]|
        |[0.11587093205757598,0.3888888888888889]|
        |[0.1139820728616421,0.3888888888888889] |
        |[0.12260126542983639,0.4722222222222222]|
        +----------------------------------------+
        only showing top 5 rows

        """

        # Check if columns argument must be a string or list datatype:
        self.__assertTypeStrOrList(columns, "columns")

        # Check if columns to be process are in dataframe
        self.__assertColsInDF(columnsProvided=columns, columnsDF=self.__df.columns)

        # Check if nameOutputCol argument a string datatype:
        self.__assertTypeStr(nameOutputCol, "nameOutpuCol")

        # Model to use vectorAssember:
        vecAssembler = VectorAssembler(inputCols=columns, outputCol="features_assembler")
        # Model for scaling feature column:
        mmScaler = MinMaxScaler(inputCol="features_assembler", outputCol=nameOutputCol)
        # Dataframe with feature_assembler column
        tempDF = vecAssembler.transform(self.__df)
        # Fitting scaler model with transformed dataframe
        model = mmScaler.fit(tempDF)

        exprs = list(filter(lambda x: x not in columns, self.__df.columns))

        exprs.extend([nameOutputCol])

        self.__df = model.transform(tempDF).select(*exprs)
        self.__addTransformation()  # checkpoint in case

        return self

    def splitStrCol(self, column, featureNames, mark):
        """This functions split a column into different ones. In the case of this method, the column provided should
        be a string of the following form 'word,foo'.

        :param column       Name of the target column, this column is going to be replaced.
        :param featureNames     List of strings of the new column names after splitting the strings.
        :param mark         String that specifies the splitting mark of the string, this frequently is ',' or ';'.
        """

        # Check if column argument is a string datatype:
        self.__assertTypeStr(column, "column")

        # Check if mark argument is a string datatype:
        self.__assertTypeStr(mark, "mark")

        assert (column in self.__df.columns), "Error: column specified does not exist in dataFrame."

        assert (type(featureNames) == type([])), "Error: featureNames must be a list of strings."

        # Setting a udf that split the string into a list of strings.
        # This is "word, foo" ----> ["word", "foo"]
        func = udf(lambda x: x.split(mark), ArrayType(StringType()))

        self.__df = self.__df.withColumn(column, func(col(column)))
        self.undoVecAssembler(column=column, featureNames=featureNames)
        self.__addTransformation()  # checkpoint in case

        return self

    def writeDFAsJson(self, path):
        p = re.sub("}\'", "}", re.sub("\'{", "{", str(self.__df.toJSON().collect())))

        with open(path, 'w') as outfile:
            # outfile.write(str(jsonCols).replace("'", "\""))
            outfile.write(p)


if __name__ == "__main__":
    pass




# Importing DataFrameTransformer library
from optimus.DfTransf import DataFrameTransformer
# Importing DataFrameAnalyzer library
from optimus.DfAnalizer import DataFrameAnalizer


class Optimus():
    """This class can be considered as a wrapper of DataFrameTransformer and
    DataFrameAnalyzer classes.
    """
    def __init__(self, df, pathFile, pu=1):
        """Constructor.
        In this method DataFrameTransformer and DataFrameAnalyzer classes are
        instantiated.
        """
        # Path file:
        self.__pathFile = pathFile

        # Pu file:
        self.__pu = pu

        # Instance of transformer class:
        self.transformer = DataFrameTransformer(df)

        # Instance of analyzer class:
        self.analyzer = DataFrameAnalizer(df, self.__pathFile, self.__pu)

    def getDataframe(self):
        """This function return the dataframe of the class"""
        return self.transformer.getDataframe()

    def setDataframe(self, df):
        """This function set a dataframe into the class for subsequent actions.
        """
        assert isinstance(df, pyspark.sql.dataframe.DataFrame), "Error: df argument must a sql.dataframe type"
        self.transformer.setDataframe(df)

    def __executeAnalyzer(self, columns):
        # First
        self.analyzer.unpersistDF()
        del self.analyzer
        # Instance of analyzer class:
        self.analyzer = DataFrameAnalizer(self.transformer.getDataframe(), self.__pathFile, self.__pu)
        # Sampling if it is specified (this method is depedent to the pu argument)
        self.analyzer.analyzeSample()
        # Analize column specified by user:
        self.analyzer.columnAnalize(columnList=columns,
                                    plots=True,
                                    valuesBar=False,
                                    printType=True,
                                    numBars=50)
    def trimCol(self, columns):
        """This methods cut left and right extra spaces in column strings provided by user.
        :param columns   list of column names of dataFrame.
                        If a string "*" is provided, the method
                        will do the trimming operation in whole dataFrame.

        :return transformer object
        """
        # Calling trimCol
        self.transformer.trimCol(columns=columns)
        # Execute analyzer:
        self.__executeAnalyzer(columns)

    def dropCol(self, df, columns):
        """This method eliminate the list of columns provided by user.
        :param columns      list of columns names or a string (a column name).

        :return transformer object
        """

        # Calling dropCol
        self.transformer.dropCol(columns=columns)
        # Execute analyzer:
        self.__executeAnalyzer(columns)

    def lowerCase(self, columns):
        """This function set all strings in columns of dataframe specified to lowercase.
        Columns argument must be a string or a list of string. In order to apply this
        function to all dataframe, columns must be equal to '*'"""

        # Calling lowerCase
        self.transformer.lowerCase(columns=columns)
        # Execute analyzer:
        self.__executeAnalyzer(columns)

    def upperCase(self, columns):
        """This function set all strings in columns of dataframe specified to uppercase.
        Columns argument must be a string or a list of string. In order to apply this function to all
        dataframe, columns must be equal to '*'"""

        # Calling lowerCase
        self.transformer.upperCase(columns=columns)
        # Execute analyzer:
        self.__executeAnalyzer(columns)

    def keepCol(self, columns):
        """This method keep only columns specified by user with columns argument in DataFrame.
        :param columns list of columns or a string (column name).

        :return transformer object
        """

        # Calling keepCol
        self.transformer.keepCol(columns=columns)
        # Execute analyzer:
        self.__executeAnalyzer(columns)
